/*! \file enc28j60.c \brief Microchip ENC28J60 Ethernet Interface Driver. */
//*****************************************************************************
//
// File Name	: 'enc28j60.c'
// Title		: Microchip ENC28J60 Ethernet Interface Driver
// Author		: Pascal Stang (c)2005
// Created		: 9/22/2005
// Revised		: 9/22/2005
// Version		: 0.1
// Target MCU	: Atmel AVR series
// Editor Tabs	: 4
//
// Description	: This driver provides initialization and transmit/receive
//	functions for the Microchip ENC28J60 10Mb Ethernet Controller and PHY.
// This chip is novel in that it is a full MAC+PHY interface all in a 28-pin
// chip, using an SPI interface to the host processor.
//
//*****************************************************************************

#include <inttypes.h>
#include <stdio.h>
#include <ctype.h>

#include "bsp.h"
#include "eth-config.h"
#include "spi.h"
#include "enc28j60.h"
#include "uip/uip.h"
#include "tty.h"
#include "led.h"

#define MIN(a,b) (a) < (b) ? (a) : (b)

uint8_t Enc28j60Bank;
uint16_t NextPacketPtr;

void pmenc_init(void)
{
  enc28j60Init();
}

void pmenc_send(void)
{
#if 0
  cprintf("pmenc send %d : ", uip_len);
  cprintf("%x %x %x\n",
	  uip_buf, 6,
	  uip_buf+6, 6,
	  uip_buf+12, 2);
#endif
  enc28j60PacketSend(uip_len, uip_buf);
}

uint16_t pmenc_read(void)
{
  uip_len = enc28j60PacketReceive(UIP_BUFSIZE, uip_buf);
#if 0
  if (uip_len)
    {
      cprintf("pmenc read %d : ", uip_len);
      cprintf("%x %x %x\n",
	      uip_buf, 6,
	      uip_buf+6, 6,
	      uip_buf+12, 2);
    }
#endif
  return uip_len;
}

void nicGetMacAddress(uint8_t* macaddr)
{
  uint8_t *p = macaddr;
	// read MAC address registers
	// NOTE: MAC address in ENC28J60 is byte-backward
	*macaddr++ = enc28j60Read(MAADR5);
	*macaddr++ = enc28j60Read(MAADR4);
	*macaddr++ = enc28j60Read(MAADR3);
	*macaddr++ = enc28j60Read(MAADR2);
	*macaddr++ = enc28j60Read(MAADR1);
	*macaddr++ = enc28j60Read(MAADR0);
	//cprintf("get mac %x\n", p, 6);
}

void nicSetMacAddress(uint8_t* macaddr)
{
	// write MAC address
	// NOTE: MAC address in ENC28J60 is byte-backward
	//cprintf("set mac %x\n", macaddr, 6);
	enc28j60Write(MAADR5, *macaddr++);
	enc28j60Write(MAADR4, *macaddr++);
	enc28j60Write(MAADR3, *macaddr++);
	enc28j60Write(MAADR2, *macaddr++);
	enc28j60Write(MAADR1, *macaddr++);
	enc28j60Write(MAADR0, *macaddr++);
}

void nicRegDump(void)
{
}


uint8_t enc28j60ReadOp(uint8_t op, uint8_t address)
{
	uint8_t data;
   
	spi_ethernet ();
	// assert CS
	ENC_CS_IO = 0;
	
	// issue read command
	spi_write (op | (address & ADDR_MASK));
	if(address & 0x80)
	{
	  spi_read ();
	}
	data = spi_read ();
	
	// release CS
	ENC_CS_IO = 1;

	return data;
}

void enc28j60WriteOp(uint8_t op, uint8_t address, uint8_t data)
{
	spi_ethernet ();
	// assert CS
	ENC_CS_IO = 0;

	// issue write command
	spi_write (op | (address & ADDR_MASK));
	// write data
	spi_write (data);

	// release CS
	ENC_CS_IO = 1;
}

void enc28j60ReadBuffer(uint16_t len, uint8_t* data)
{
	spi_ethernet ();
	// assert CS
	ENC_CS_IO = 0;
	
	// issue read command
	spi_write (ENC28J60_READ_BUF_MEM);
	while(len--)
	{
		// read data
		*data++ = spi_read ();
	}	
	// release CS
	ENC_CS_IO = 1;
}

void enc28j60WriteBuffer(uint16_t len, uint8_t* data)
{
	spi_ethernet ();
	// assert CS
	ENC_CS_IO = 0;
	
	// issue write command
	spi_write (ENC28J60_WRITE_BUF_MEM);
	while(len--)
	{
		// write data
		spi_write (*data++);
	}	
	// release CS
	ENC_CS_IO = 1;
}

void enc28j60SetBank(uint8_t address)
{
	// set the bank (if needed)
	if((address & BANK_MASK) != Enc28j60Bank)
	{
		// set the bank
		enc28j60WriteOp(ENC28J60_BIT_FIELD_CLR, ECON1, (ECON1_BSEL1|ECON1_BSEL0));
		enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, (address & BANK_MASK)>>5);
		Enc28j60Bank = (address & BANK_MASK);
	}
}

uint8_t enc28j60Read(uint8_t address)
{
	uint8_t i, j;
	// set the bank
	enc28j60SetBank(address);
	// do the read
	i = enc28j60ReadOp(ENC28J60_READ_CTRL_REG, address);
	if (address & 0x80)
	  {
	    do {
	      j = i;
	      i = enc28j60ReadOp(ENC28J60_READ_CTRL_REG, address);
	      if (i != j)
		cprintf("%b : %b %b ??\n", address, i, j);
	    } while (i != j);
	  }
	return i;
}

void enc28j60Write(uint8_t address, uint8_t data)
{
	// set the bank
	enc28j60SetBank(address);
	// do the write
	enc28j60WriteOp(ENC28J60_WRITE_CTRL_REG, address, data);

	if (address & 0x80)
	  {
	    uint8_t j = enc28j60ReadOp(ENC28J60_READ_CTRL_REG, address);
	    if (j != data)
	      cprintf("%b : %b %b !!\n", address, data, j);
	  }
}

uint16_t enc28j60PhyRead(uint8_t address)
{
	uint16_t data;

	// Set the right address and start the register read operation
	enc28j60Write(MIREGADR, address);
	enc28j60Write(MICMD, MICMD_MIIRD);

	// wait until the PHY read completes
	while(enc28j60Read(MISTAT) & MISTAT_BUSY);

	// quit reading
	enc28j60Write(MICMD, 0x00);
	
	// get data value
	data  = enc28j60Read(MIRDH);
	data <<= 8;
	data |= enc28j60Read(MIRDL);

	// return the data
	return data;
}

void enc28j60PhyWrite(uint8_t address, uint16_t data)
{
	// set the PHY register address
	enc28j60Write(MIREGADR, address);
	
	// write the PHY data
	enc28j60Write(MIWRL, data);	
	enc28j60Write(MIWRH, data>>8);

	// wait until the PHY write completes
	while(enc28j60Read(MISTAT) & MISTAT_BUSY);
}

void enc28j60Init(void)
{
	spi_ethernet ();

	// perform system reset
	enc28j60WriteOp(ENC28J60_SOFT_RESET, 0, ENC28J60_SOFT_RESET);
	wait_ms(1);
	// check CLKRDY bit to see if reset is complete
	//while(!(enc28j60Read(ESTAT) & ESTAT_CLKRDY));
	// Errata workaround #2, CLKRDY check is unreliable, delay 1 mS instead
	wait_ms(5);

	// lamp test
//	enc28j60PhyWrite(PHLCON, 0x0AA2);



	// do bank 0 stuff
	// initialize receive buffer
	// 16-bit transfers, must write low byte first

	// set receive buffer start address
	NextPacketPtr = RXSTART_INIT;
	enc28j60Write(ERXSTL, RXSTART_INIT&0xFF);
	enc28j60Write(ERXSTH, RXSTART_INIT>>8);

	// set receive pointer address
	enc28j60Write(ERXRDPTL, RXSTART_INIT&0xFF);
	enc28j60Write(ERXRDPTH, RXSTART_INIT>>8);

	// set receive buffer end
	// ERXND defaults to 0x1FFF (end of ram)
	enc28j60Write(ERXNDL, RXSTOP_INIT&0xFF);
	enc28j60Write(ERXNDH, RXSTOP_INIT>>8);

	// set transmit buffer start
	// ETXST defaults to 0x0000 (beginnging of ram)
	enc28j60Write(ETXSTL, TXSTART_INIT&0xFF);
	enc28j60Write(ETXSTH, TXSTART_INIT>>8);

	// do bank 2 stuff
	// enable MAC receive
	enc28j60Write(MACON1, MACON1_MARXEN|MACON1_TXPAUS|MACON1_RXPAUS);
	// bring MAC out of reset
	enc28j60Write(MACON2, 0x00);
	// enable automatic padding and CRC operations
	enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, MACON3, MACON3_PADCFG0|MACON3_TXCRCEN|MACON3_FRMLNEN);
//	enc28j60Write(MACON3, MACON3_PADCFG0|MACON3_TXCRCEN|MACON3_FRMLNEN);
	// set inter-frame gap (non-back-to-back)
	enc28j60Write(MAIPGL, 0x12);
	enc28j60Write(MAIPGH, 0x0C);
	// set inter-frame gap (back-to-back)
	enc28j60Write(MABBIPG, 0x12);
	// Set the maximum packet size which the controller will accept
	enc28j60Write(MAMXFLL, MAX_FRAMELEN&0xFF);	
	enc28j60Write(MAMXFLH, MAX_FRAMELEN>>8);

	enc28j60Write(ERXFCON, ERXFCON_UCEN|ERXFCON_BCEN|ERXFCON_CRCEN);

	// do bank 3 stuff
	// write MAC address
	// NOTE: MAC address in ENC28J60 is byte-backward
	enc28j60Write(MAADR5, ENC28J60_MAC0);
	enc28j60Write(MAADR4, ENC28J60_MAC1);
	enc28j60Write(MAADR3, ENC28J60_MAC2);
	enc28j60Write(MAADR2, ENC28J60_MAC3);
	enc28j60Write(MAADR1, ENC28J60_MAC4);
	enc28j60Write(MAADR0, ENC28J60_MAC5);

	// no loopback of transmitted frames
	enc28j60PhyWrite(PHCON2, PHCON2_HDLDIS);

	// switch to bank 0
	enc28j60SetBank(ECON1);
	// enable interrutps
	enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, EIE, EIE_INTIE|EIE_PKTIE);
	// enable packet reception
	enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, ECON1_RXEN);

	enc28j60PhyWrite(PHLCON, 0x0126);

	//cprintf("rev: %d\n", enc28j60Read(EREVID));
/*
	// setup duplex ----------------------

	// Disable receive logic and abort any packets currently being transmitted
	enc28j60WriteOp(ENC28J60_BIT_FIELD_CLR, ECON1, ECON1_TXRTS|ECON1_RXEN);
	
	{
		uint16_t temp;
		// Set the PHY to the proper duplex mode
		temp = enc28j60PhyRead(PHCON1);
		temp &= ~PHCON1_PDPXMD;
		enc28j60PhyWrite(PHCON1, temp);
		// Set the MAC to the proper duplex mode
		temp = enc28j60Read(MACON3);
		temp &= ~MACON3_FULDPX;
		enc28j60Write(MACON3, temp);
	}

	// Set the back-to-back inter-packet gap time to IEEE specified 
	// requirements.  The meaning of the MABBIPG value changes with the duplex
	// state, so it must be updated in this function.
	// In full duplex, 0x15 represents 9.6us; 0x12 is 9.6us in half duplex
	//enc28j60Write(MABBIPG, DuplexState ? 0x15 : 0x12);	
	
	// Reenable receive logic
	enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, ECON1_RXEN);

	// setup duplex ----------------------
*/
}


#define ETHERNET_MIN_PACKET_LENGTH	0x3C
#define ETHERNET_HEADER_LENGTH		0x0E

#define IP_TCP_HEADER_LENGTH 40
#define TOTAL_HEADER_LENGTH (IP_TCP_HEADER_LENGTH+ETHERNET_HEADER_LENGTH)


void enc28j60PacketSend(uint16_t len, uint8_t* packet)
{

/*
	// dump packet
	int i,j;
	for (i=0;i<len;i+=16)
	{
		printf("%04x  ",i);
		for (j=0;j<16;j++)
			printf("%02x ",(i+j) < len ? packet[i+j] : 0);
		printf("  ");
		for (j=0;j<16;j++)
			printf("%c", (i+j) < len ? isprint(packet[i+j]) ? packet[i+j] : '.' : '.');
 		printf("\n");
	}
*/

	// Set the write pointer to start of transmit buffer area
	enc28j60Write(EWRPTL, TXSTART_INIT);
	enc28j60Write(EWRPTH, TXSTART_INIT>>8);

	// Set the TXND pointer to correspond to the packet size given
	enc28j60Write(ETXNDL, (TXSTART_INIT+len));
	enc28j60Write(ETXNDH, (TXSTART_INIT+len)>>8);

	// write per-packet control byte
	enc28j60WriteOp(ENC28J60_WRITE_BUF_MEM, 0, 0x00);

	// TODO, fix this up

	if( uip_len <= TOTAL_HEADER_LENGTH )
	{
		// copy the packet into the transmit buffer
		enc28j60WriteBuffer(len, packet);
	}
	else
	{
      	len -= TOTAL_HEADER_LENGTH;
		enc28j60WriteBuffer(TOTAL_HEADER_LENGTH, packet);
		enc28j60WriteBuffer(len, (unsigned char *)uip_appdata);
	}



	//
	// Errata workaround #12, reset TX logic
	//
	{
		// clear TXRST bit in ECON1
		//enc28j60WriteOp(ENC28J60_BIT_FIELD_CLR, ECON1, ECON1_TXRST);
		//delay(100);
		// set TXRST bit in ECON1
		//enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, ECON1_TXRST);
	}
	// send the contents of the transmit buffer onto the network
	enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON1, ECON1_TXRTS);
}




uint16_t enc28j60PacketReceive(uint16_t maxlen, uint8_t* packet)
{
	uint16_t rxstat;
	uint16_t len;

	// check if a packet has been received and buffered
	if( !(enc28j60Read(EIR) & EIR_PKTIF) )
	{
		// Errata workaround #6, PKTIF is not reliable
		// double check by looking at EPKTCNT
		if (enc28j60Read(EPKTCNT) == 0)
			return 0;
	}
	
	// Make absolutely certain that any previous packet was discarded	
	//if( WasDiscarded == FALSE)
	//	MACDiscardRx();

	// Set the read pointer to the start of the received packet
	enc28j60Write(ERDPTL, (NextPacketPtr));
	enc28j60Write(ERDPTH, (NextPacketPtr)>>8);

	// read the next packet pointer
	NextPacketPtr  = enc28j60ReadOp(ENC28J60_READ_BUF_MEM, 0);
	NextPacketPtr |= enc28j60ReadOp(ENC28J60_READ_BUF_MEM, 0)<<8;

	// read the packet length
	len  = enc28j60ReadOp(ENC28J60_READ_BUF_MEM, 0);
	len |= enc28j60ReadOp(ENC28J60_READ_BUF_MEM, 0)<<8;

	// read the receive status
	rxstat  = enc28j60ReadOp(ENC28J60_READ_BUF_MEM, 0);
	rxstat |= enc28j60ReadOp(ENC28J60_READ_BUF_MEM, 0)<<8;

	// limit retrieve length
	// (we reduce the MAC-reported length by 4 to remove the CRC)
	len = MIN(len, maxlen);

	// copy the packet from the receive buffer
	enc28j60ReadBuffer(len, packet);

	// Move the RX read pointer to the start of the next received packet
	// This frees the memory we just read out
/*
	enc28j60Write(ERXRDPTL, (NextPacketPtr));
	enc28j60Write(ERXRDPTH, (NextPacketPtr)>>8);
*/
	// Errata workaround #13. Make sure ERXRDPT is odd
	//
	{
		uint16_t rs,re;
		rs = enc28j60Read(ERXSTH);
		rs <<= 8;
		rs |= enc28j60Read(ERXSTL);
		re = enc28j60Read(ERXNDH);
		re <<= 8;
		re |= enc28j60Read(ERXNDL);
		if (NextPacketPtr - 1 < rs || NextPacketPtr - 1 > re)
		{
			enc28j60Write(ERXRDPTL, (re));
			enc28j60Write(ERXRDPTH, (re)>>8);
		}
		else
		{
			enc28j60Write(ERXRDPTL, (NextPacketPtr-1));
			enc28j60Write(ERXRDPTH, (NextPacketPtr-1)>>8);
		}
	}

	// decrement the packet counter indicate we are done with this packet
	enc28j60WriteOp(ENC28J60_BIT_FIELD_SET, ECON2, ECON2_PKTDEC);

	return len;
}

void enc28j60ReceiveOverflowRecover(void)
{
	// receive buffer overflow handling procedure

	// recovery completed
}
